// avisynth redirecter dll
// MobileHackerz http://www.nurs.or.jp/~calcium/

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include "internal.h"

#include <windows.h>


typedef __int64 int64_t;
#include "avisynthdll.h"

extern "C" {
__declspec(dllexport) void * __stdcall avs_create();
__declspec(dllexport) int __stdcall avs_open(void *avs, char *fname, AVSDLLVideoInfo *vi);
__declspec(dllexport) int __stdcall avs_getvframe(void *avs, int clip_num, int frm, AVSDLLVideoFrame *vf);
__declspec(dllexport) int __stdcall avs_getaframe(void *avs, int clip_num, void *buf, AVSDLLAudioPos *pos);
__declspec(dllexport) int __stdcall avs_close(void *avs, int clip_num);
__declspec(dllexport) int __stdcall avs_getlasterror(char *str,int len);
}


#define MAX_CLIPS  10
#define ERRMSG_LEN 1024

static PClip	clip[MAX_CLIPS];
static int		g_clip_flg[MAX_CLIPS];

static char g_lasterr[ERRMSG_LEN];

// -----------------------------------------------------------------------------------------------------------

void * __stdcall avs_create()
{
	HMODULE avsdll = LoadLibrary("avisynth.dll");
	if(!avsdll) { 
		strncpy(g_lasterr,"Cannot load avisynth.dll",ERRMSG_LEN-1);
		return NULL;
	}
	IScriptEnvironment* (* CreateScriptEnvironment)(int version) = (IScriptEnvironment*(*)(int)) GetProcAddress(avsdll, "CreateScriptEnvironment");
	if(!CreateScriptEnvironment) { 
		strncpy(g_lasterr,"Cannot load CreateScriptEnvironment",ERRMSG_LEN-1);
		return NULL;
	}
	IScriptEnvironment* env = CreateScriptEnvironment(AVISYNTH_INTERFACE_VERSION);
	if (env == NULL) {
		strncpy(g_lasterr,"Required Avisynth 2.5",ERRMSG_LEN-1);
	} else {
		g_lasterr[0] = 0;
	}
	return env;
}

int __stdcall avs_open(void *avs, char *fname, AVSDLLVideoInfo *vi)
{
	IScriptEnvironment* env = (IScriptEnvironment*)avs;
	int clip_num;
	int i;

	try {
		AVSValue arg(fname);
		AVSValue res = env->Invoke("Import", AVSValue(&arg, 1));
		if (!res.IsClip()) {
			strncpy(g_lasterr,"The script's return was not a video clip.",ERRMSG_LEN-1);
			return -1;
		}

		clip_num = -1;
		for(i=0;i<MAX_CLIPS;i++) {
			if (g_clip_flg[i] == 0) {
				clip_num = i;
				g_clip_flg[i] = 1;
				break;
			}
		}
		if (clip_num == -1) {
			strncpy(g_lasterr,"Too many open clips",ERRMSG_LEN-1);
			return -1;
		}

		clip[clip_num] = res.AsClip();
		VideoInfo inf  = clip[clip_num]->GetVideoInfo();

		if (inf.HasVideo()) {
			if(!inf.IsYV12()) {
				res = env->Invoke("ConvertToYV12", AVSValue(&res, 1));
				clip[clip_num] = res.AsClip();
				inf = clip[clip_num]->GetVideoInfo();
			}
			if(!inf.IsYV12()) {
				strncpy(g_lasterr,"Cannot convert video to YUV2",ERRMSG_LEN-1);
				return -1;
			}
			if(inf.IsFieldBased()) {
				strncpy(g_lasterr,"Sorry interlaced clip not supoorted",ERRMSG_LEN-1);
				return -1;
			}
		}

		if (inf.HasAudio()) {
			if(inf.SampleType() != SAMPLE_INT16) {
				res = env->Invoke("ConvertAudioTo16bit", res);
				clip[clip_num] = res.AsClip();
				inf = clip[clip_num]->GetVideoInfo();
				if(inf.SampleType() != SAMPLE_INT16) {
					strncpy(g_lasterr,"Cannot convert audio to 16bit",ERRMSG_LEN-1);
					return -1;
				}
			}
		}

		if (vi != NULL) {
			vi->width   = inf.width;
			vi->height  = inf.height;
			vi->raten   = inf.fps_numerator;
			vi->rated   = inf.fps_denominator;
			vi->aspectn = 0;
			vi->aspectd = 1;
			vi->interlaced_frame = 0; // progressive
			vi->top_field_first  = 0;
			vi->num_frames = inf.num_frames;
	
			vi->audio_samples_per_second = inf.SamplesPerSecond();
			vi->num_audio_samples        = inf.num_audio_samples;
			vi->sample_type              = inf.SampleType();
			vi->nchannels                = inf.AudioChannels();
		}

		g_lasterr[0] = 0;
		return clip_num;
	} catch(AvisynthError err) {
		strncpy(g_lasterr,err.msg,ERRMSG_LEN-1);	
		return -1;
	}
}

int __stdcall avs_getvframe(void *avs, int clip_num, int frm, AVSDLLVideoFrame *vf)
{
	IScriptEnvironment* env = (IScriptEnvironment*)avs;

	try {
		VideoInfo inf = clip[clip_num]->GetVideoInfo();
		PVideoFrame f = clip[clip_num]->GetFrame(frm, env);
	
		static const int planes[] = {PLANAR_Y, PLANAR_U, PLANAR_V};

		for(int p=0; p<3; p++) {
			vf->plane[p].width  = inf.width  >> (p ? 1 : 0);
			vf->plane[p].height = inf.height >> (p ? 1 : 0);
			vf->plane[p].pitch  = f->GetPitch(planes[p]);
			vf->plane[p].data   = (BYTE *)f->GetReadPtr(planes[p]);
		}
		g_lasterr[0] = 0;
		return 0;
	} catch(AvisynthError err) {
		strncpy(g_lasterr,err.msg,ERRMSG_LEN-1);	
		return -1;
	}

}

int __stdcall avs_getaframe(void *avs, int clip_num, void *buf, AVSDLLAudioPos *pos)
{
	IScriptEnvironment* env = (IScriptEnvironment*)avs;

	try {
		clip[clip_num]->GetAudio(buf,pos->start,pos->count,env);
		g_lasterr[0] = 0;
		return 0;
	} catch(AvisynthError err) {
		strncpy(g_lasterr,err.msg,ERRMSG_LEN-1);	
		return -1;
	}
}

int __stdcall avs_close(void *avs, int clip_num)
{
	delete clip[clip_num];
	g_clip_flg[clip_num] = 0;
	g_lasterr[0] = 0;
	return 0;
}

int __stdcall avs_getlasterror(char *str,int len)
{
	return (int)strlen(strncpy(str,g_lasterr,len-1));
}

// -----------------------------------------------------------------------------------------------------------

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	UNREFERENCED_PARAMETER(hinstDLL);
	UNREFERENCED_PARAMETER(lpvReserved);

	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			for(int i=0;i<MAX_CLIPS;i++) { g_clip_flg[i] = 0; }
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}

